module Labyrinth
module QCutList
module CutList
module PartsIntelligence
  def layoutImpossible?
    return true if @inPartList.empty?
    return true if !@homerDepot.boardsAvailable?
  end
  def allPartsSameThickness?
    @inPartList.allPartsSameThickness?
  end
  def allPartsSameMaterial?
    @inPartList.allPartsSameMaterial?
  end
  def partsWiderThanBoards?
    @widestBoard = @homerDepot.widestBoardAvailable?
    partArray = @inPartList.getList
    partArray.each { |part| return true if part.getWidth > @widestBoard }
    return false
  end
  def partsThickerThanBoards?
    @thickestBoard = @homerDepot.thickestBoardAvailable?.inch
    partArray = @inPartList.getList
    partArray.each { |part| return true if part.getThickness > @thickestBoard }
    return false
  end
  def partsThinnerThanBoards?
    @thickestBoard = @homerDepot.thickestBoardAvailable?.inch
    partArray = @inPartList.getList
    partArray.each { |part| return true if part.getThickness < @thickestBoard }
    return false
  end
  def partsMutipleofWidth?
  end
  def morePartsThanBoards?
    return false
  end
end #module PartsIntelligence
class LayoutPreParser
  include PartsIntelligence
  def initialize(partsList, inBoardList, partOptions,layoutOptions,metricVolume)
    @homerDepot = BoardSuperMarket.new(inBoardList, partOptions,metricVolume)
    @layoutOptions = layoutOptions
    @layoutByMaterial = layoutOptions[:layoutByMaterial] 
    @quarters = partOptions[2].include? true
    @inPartList = partsList
    @status = LayoutStatus.new("","success")
    @partsIgnoreList = CutListPartList.new
  end
  def partsPreParser
    if layoutImpossible?
      @status.set("No parts or no boards.\nLayout is not possible","error")
      return
    end
    if morePartsThanBoards?
      @status.set("Parts exceeds available boards.\nLayout is not possible","error")
      return
    end
    if ( partsThickerThanBoards? || partsThinnerThanBoards? )&& @quarters
      if @layoutOptions[:splitThickParts]
        normalizeThinGlueUpParts(@thickestBoard)
        normalizeGlueUpParts(@thickestBoard)
      else
        @status.set("Some parts require thicker boards than is available.\n" + \
                          "Either check the 'Split Thick Parts' option or specify" + \
                          " thicker boards\n" + \
                          "Some or all of the parts will not be layed out","warning")
        removeThickParts(@thickestBoard)
      end
    end
    if !allPartsSameMaterial?
    end
    if partsWiderThanBoards?
      normalizePartsList(@widestBoard) if @layoutOptions[:splitWideParts]
    end
  end
  def removeThickParts(thickestBoard)
    partArray = @inPartList.getList
    removalList = Array.new
    partArray.each { |part| 
      if part.getThickness.inch > thickestBoard.inch
        removalList.push(part)
      end
    }
    removalList.each{ |oldPart| 
      @inPartList.remove(oldPart)
      @partsIgnoreList.add(oldPart)
     }
end
  def normalizeGlueUpParts(thickestBoard)
    partArray = @inPartList.getList
    removalList = Array.new
    addList = Array.new
    partArray.each { |part| 
      if part.getThickness.inch > thickestBoard.inch
        splitInto = (part.getThickness.inch/thickestBoard.inch).ceil
        if splitInto >= 10
          @status.set("Part " + "'" + part.getName + "'" + " is too thick" + "\nThis part will not be split and will be ignored in the layout ","warning")
          removalList.push(part)
          @partsIgnoreList.add(part)
          next
        end
        splitThickness = thickestBoard.inch
        splitInto.times do |i| 
          splitPart = part.deep_clone
          splitPart.changeThickness(splitThickness)
          addList.push(splitPart)
        end
        removalList.push(part)
      end
    }
    removalList.each{ |oldPart| @inPartList.remove(oldPart) }
    addList.each{ |newPart| @inPartList.add(newPart)}
    @inPartList.sort
  end
  def normalizeThinGlueUpParts(thickestBoard)
    partArray = @inPartList.getList
    removalList = Array.new
    addList = Array.new
    partArray.each { |part| 
      if part.getThickness.inch < thickestBoard.inch
        removalList.push(part)
        newPart = part.deep_clone
        newPart.changeThickness(part.getMarginThickness)
        addList.push(newPart)
      end
    }
    removalList.each{ |oldPart| @inPartList.remove(oldPart) }
    addList.each{ |newPart| @inPartList.add(newPart)}
    @inPartList.sort
  end
  def normalizePartsList(widestBoard)
    partArray = @inPartList.getList
    removalList = Array.new
    addList = Array.new
    partArray.each { |part| 
      if part.getWidth.inch> widestBoard.inch
          splitInto = (part.getWidth.inch/widestBoard.inch).ceil
          splitWidth = widestBoard.inch
          splitInto.times do |i| 
            splitPart = part.deep_clone
            splitPart.changeWidth(splitWidth)
            addList.push(splitPart)
          end
          removalList.push(part)
      end
      }
    removalList.each{ |oldPart| @inPartList.remove(oldPart) }
    addList.each{ |newPart| @inPartList.add(newPart)}
    @inPartList.sort
  end
  def partsSplitter
    if @layoutByMaterial  
        @listOfPartsLists = @inPartList.splitPartsListByThicknessAndMaterial
    else
        @listOfPartsLists = @inPartList.splitPartsListByThickness
    end
  end
  def run
    partsPreParser()
    return nil if @status.statusNotGood
    partsSplitter()
    return nil if @status.statusNotGood
    return @listOfPartsLists, @partsIgnoreList
  end
end
class SolidPartPreParser < LayoutPreParser
end
class SheetPartPreParser < LayoutPreParser
end
class LayoutEngine  
  include PartsIntelligence
  def initialize(inPartList, inBoardList, boardOptions,layoutOptions,metricVolume,
  	availableSheets=nil, edgeWaste=0)
    @homerDepot = BoardSuperMarket.new(inBoardList, boardOptions,metricVolume, availableSheets, edgeWaste)
    @inPartList = inPartList
    @inPartList.sort
    @originalInPartList = inPartList.deep_clone
    @inBoardList = inBoardList
    @metricVolume = metricVolume
    @layoutBoards = Array.new
    @status = LayoutStatus.new("","success")
    @partsToDoList = CutListPartList.new
    @partsIgnoreList = CutListPartList.new
    @convergenceImpossible = false
    @layoutOptions = layoutOptions
    @quarters = boardOptions[2]
    @thicknessOptions = (@quarters.include? true )
  end
  def convergenceImpossible?
     return @convergenceImpossible == true
  end
  def getBoard
      thickness = @inPartList.getList[0].getThickness 
    if @layoutOptions[:layoutByMaterial]
      material = @inPartList.getList[0].getMaterial
    else
      material = "generic"
    end
    @board = @homerDepot.getBoard(material, thickness)
    return @board
  end
  def _getBoard
  end
  def placeParts
    @remainingPartsToPlace = @inPartList.count
    board = getBoard()
    if board != nil
      layoutBoard = LayoutBoard.new(board,@layoutOptions,@metricVolume)
      @layoutBoards.push(layoutBoard)
      while !@inPartList.empty?
        nextPart = @inPartList.removeFirst!
        if ( layoutBoard.addPartToBoard(nextPart) == false )
          @partsToDoList.add(nextPart)
        else
        end
      end
      tempListPtr = @inPartList
      @inPartList = @partsToDoList
      @partsToDoList = tempListPtr
      @inPartList.sort
      if @remainingPartsToPlace == @inPartList.count
        @status.set("Some parts were left which could not be placed on given stock.\nTry larger sizes.","warning")
        @convergenceImpossible = true
        @layoutBoards.pop()
        return
      else
        layoutBoard.setBoardFull
      end      
    end
  end
  def run
    while (@homerDepot.boardsAvailable? && !@inPartList.empty? && !convergenceImpossible?)
      placeParts
      return nil if @status.statusNotGood
    end
    if (!@homerDepot.boardsAvailable? && !@inPartList.empty?)
      @status.set("Some parts were left which could not be placed.\nNo boards left or board limit reached.","warning")
    end
    return nil if @status.statusNotGood
    return nil, @inPartList if @layoutBoards.empty?
    return @layoutBoards, @inPartList
  end
end
class BestFitLayoutEngine < LayoutEngine
end
end # CutList
end # QCutList
end # Labyrinth
